//---------------------------------------------------------------------------
// JTHZ backup deluxe
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <inifiles.hpp>
#include "unit1.h"
//---------------------------------------------------------------------------
USERES("budl.res");
USEFORM("Unit1.cpp", Form1);
//---------------------------------------------------------------------------
#define FOUND_SUCCESS 0
//---------------------------------------------------------------------------
bool maskMatch(String,String);
bool match(const char*,const char*);
bool DirectoryExists(String);
void ForceDirectories(String);
void sweep(String,bool,void(*)(String,TSearchRec,bool));
void p_new(String,TSearchRec,bool);
void p_obs(String,TSearchRec,bool);
void p_dirKiller(String,TSearchRec,bool);
bool skip(String);
bool keepAnyway(String);
void err(String);
//---------------------------------------------------------------------------
TStringList *recursed=0,*non_recursed=0,*include=0;
TStringList *exclude=0,*dirHitList=0,*keep=0;
String destination;
//---------------------------------------------------------------------------
bool maskMatch(String mask,String fileName){
    // returns true if fileName matches standard dos wildcard mask
    // derived from Bob Stout's XSTRCMP.C
    return match(mask.LowerCase().c_str(),fileName.LowerCase().c_str());
}
//---------------------------------------------------------------------------
bool match(const char *mask,const char *s){
    // recursive helperfunction for maskMatch
    switch(*mask){
        case '\0':
            return !*s;
        case '*':
            return match(mask+1,s)||*s&&match(mask,s+1);
        case '?':
            return *s&&match(mask+1,s+1);
    }
    return(*mask==*s)&&match(mask+1,s+1);
}
//---------------------------------------------------------------------------
bool DirectoryExists(String Dir){
    // hz translation from Delphi/filectrl original
    // to avoid overhead of linking in vclx40.lib (247K)
    int Code=GetFileAttributes(Dir.c_str());
    return (Code!=-1)&&(FILE_ATTRIBUTE_DIRECTORY&Code);
}
//---------------------------------------------------------------------------
void ForceDirectories(String Dir){
    // hz translation from Delphi/filectrl original
    // to avoid overhead of linking in vclx40.lib (247K)
    if(!Dir.Length())
        return;
    if(*AnsiLastChar(Dir)=='\\')
        Dir.Delete(Dir.Length(),1);
    if((Dir.Length()<3)||DirectoryExists(Dir)||(ExtractFilePath(Dir)==Dir))
        return;
    ForceDirectories(ExtractFilePath(Dir));
    CreateDir(Dir);
}
//---------------------------------------------------------------------------
void sweep(String inputDir,bool bRecurse,
        void(*generic)(String,TSearchRec,bool)){
    // recurses directory if bRecurse==true
    if(!inputDir.Length())
        return;
    if(*AnsiLastChar(inputDir)!='\\')
        inputDir+="\\";
    TSearchRec sr;
    if(FindFirst(inputDir+"*.*",faAnyFile,sr)==FOUND_SUCCESS){
        generic(inputDir,sr,bRecurse);
        while(FindNext(sr)==FOUND_SUCCESS)
            generic(inputDir,sr,bRecurse);
        FindClose(sr);
    }
}
//---------------------------------------------------------------------------
void p_new(String path,TSearchRec tsr,bool bRecurse){
    // sweep plugin-function to hunt for new files
    if(tsr.Attr&faDirectory){
        if(bRecurse&&tsr.Name[1]!='.')
            sweep(path+tsr.Name,true,p_new);
    }
    else if(!skip(tsr.Name)){
        // build copy-destination (global 'destination' ends with \)
        String destFile=destination+path[1]+"\\"+
                path.SubString(4,path.Length()-3)+tsr.Name;
        TSearchRec srDest;
        if(FindFirst(destFile,faAnyFile,srDest)==FOUND_SUCCESS){
            // file exists; is our file newer?
            // use y2k-proof 64bit arithmetic     =8)~
            LARGE_INTEGER sourceTimeDate,destTimeDate;
            CopyMemory(&sourceTimeDate,&tsr.FindData.ftLastWriteTime,
                    sizeof(LARGE_INTEGER));
            CopyMemory(&destTimeDate,&srDest.FindData.ftLastWriteTime,
                    sizeof(LARGE_INTEGER));
            if(sourceTimeDate.QuadPart>destTimeDate.QuadPart){
                // remove RHSA
                SetFileAttributes(destFile.c_str(),FILE_ATTRIBUTE_NORMAL);
                CopyFile((path+tsr.Name).c_str(),destFile.c_str(),FALSE);
            }
            FindClose(srDest);
        }
        else{
            ForceDirectories(ExtractFilePath(destFile));
            CopyFile((path+tsr.Name).c_str(),destFile.c_str(),FALSE);
        }
    }
}
//---------------------------------------------------------------------------
void p_obs(String path,TSearchRec tsr,bool bRecurse){
    // sweep plugin-function to hunt for obsolete files
    if(tsr.Attr&faDirectory){
        if(bRecurse&&tsr.Name[1]!='.')
            sweep(path+tsr.Name,true,p_obs);
    }
    else{
        // map destination-path back to source-path
        String destFile=path+tsr.Name;
        int len=destination.Length();
        String checkFile=destFile.SubString(len,destFile.Length()-len+1);
        checkFile[1]=destFile[len+1];
        checkFile[2]=':';
        checkFile[3]='\\';
        if(!FileExists(checkFile)||skip(tsr.Name)){
            if(!keepAnyway(tsr.Name)){
                // remove RHSA
                SetFileAttributes(destFile.c_str(),FILE_ATTRIBUTE_NORMAL);
                DeleteFile(destFile.c_str());
            }
        }
    }
}
//---------------------------------------------------------------------------
void p_dirKiller(String path,TSearchRec tsr,bool bRecurse){
    // sweep plugin-function to hunt for obsolete directories
    if(tsr.Attr&faDirectory){
        if(bRecurse&&tsr.Name[1]!='.'){
            dirHitList->Add(path+tsr.Name);
            sweep(path+tsr.Name,true,p_dirKiller);
        }
    }
}
//---------------------------------------------------------------------------
bool skip(String fileName){
    // returns true if fileName should be skipped
    // based on entries in global TStringLists exclude and include
    // standard dos wildcards * and ? apply
    if(!fileName.Length())
        return true; // skip
    for(int i=0;i<exclude->Count;++i)
        if(maskMatch(exclude->Strings[i],fileName))
            return true; // skip
    if(include->Count){ // file must be included here to pass
        for(int i=0;i<include->Count;++i)
            if(maskMatch(include->Strings[i],fileName))
                return false; // go ahead
        return true; // skip
    }
    return false; // go ahead
}
//---------------------------------------------------------------------------
bool keepAnyway(String fileName){
    // returns true if fileName should be skipped (not obsolete deleted)
    // based on entries in global TStringList keep
    // standard dos wildcards * and ? apply
    if(!fileName.Length())
        return true; // skip
    for(int i=0;i<keep->Count;++i)
        if(maskMatch(keep->Strings[i],fileName))
            return true; // keep anyway
    return false; // nope, throw away
}
//---------------------------------------------------------------------------
void err(HWND hWnd,String errStr){
    MessageBox(hWnd,errStr.c_str(),"SORRY",MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------
#pragma argsused
WINAPI WinMain(HINSTANCE hI1,HINSTANCE hI2,PSTR pstrCmdLine,int){
    const String INI_SECTION="jthz";
    const char progid[]="JTHZ backup";
    if(OpenMutex(0,false,progid))
        return 0;
    HANDLE mutex=CreateMutex(NULL,true,progid);
    try{
        Application->Initialize();
        TForm1 *Form1=new TForm1(Application);
        Form1->Show();
        Form1->FormStyle=fsStayOnTop;
        Form1->Update();
        Application->Title="JTHZ Backup";
        Application->Run();
        recursed=new TStringList(); // recurse these directories
        non_recursed=new TStringList(); // just scan these directories
        exclude=new TStringList(); // exclude these files
        include=new TStringList(); // include just these files
        keep=new TStringList(); // leave these files when they get obsolete
        String cmdLine=String(pstrCmdLine);
        if(!cmdLine.Length()){
            err(Form1->Handle,"Run me with a parameter");
            goto THE_END;
        }
        // try to read info from file
        TIniFile *ini=new TIniFile(cmdLine);
        recursed->CommaText=ini->ReadString(INI_SECTION,"recursed","");
        non_recursed->CommaText=ini->ReadString(INI_SECTION,"non-recursed","");
        destination=ini->ReadString(INI_SECTION,"destiny","");
        exclude->CommaText=ini->ReadString(INI_SECTION,"xclude","");
        include->CommaText=ini->ReadString(INI_SECTION,"include","");
        keep->CommaText=ini->ReadString(INI_SECTION,"keep on destiny","");
        delete ini;
        // check validity of input
        if(!destination.Length()){
            err(Form1->Handle,"No destination scanned");
            goto THE_END;
        }
        if(*AnsiLastChar(destination)!='\\')
            destination+="\\";
        // make sure destination exists
        ForceDirectories(destination);
        // now see if we have access to destination
        if(!DirectoryExists(destination)){
            err(Form1->Handle,"Invalid destination");
            goto THE_END;
        }
        // source(s) valid?
        if(!recursed->Count&&!non_recursed->Count){
            err(Form1->Handle,"No source-directories scanned");
            goto THE_END;
        }
        for(int i=0;i<recursed->Count;++i)
            if(!DirectoryExists(recursed->Strings[i])){
                err(Form1->Handle,"Illegal directory in recursed dirs");
                goto THE_END;
            }
        for(int i=0;i<non_recursed->Count;++i)
            if(!DirectoryExists(non_recursed->Strings[i])){
                err(Form1->Handle,"Illegal directory in non-recursed entry");
                goto THE_END;
            }
        // first round; copy new files from source to destination
        Form1->Animate1->CommonAVI=aviCopyFile;
        Form1->Animate1->Active=true;
        Form1->Caption="Copying new files...";
        for(int i=0;i<recursed->Count;++i)
            sweep(recursed->Strings[i],true,p_new);
        for(int i=0;i<non_recursed->Count;++i)
            sweep(non_recursed->Strings[i],false,p_new);
        // second round; remove obsolete files
        Form1->Caption="Removing obsolete files...";
        Form1->Animate1->CommonAVI=aviDeleteFile;
        Form1->Animate1->Active=true;
        for(int i=0;i<recursed->Count;++i){
            String tmp=recursed->Strings[i];
            if(!tmp.Length())
                continue;
            sweep(destination+tmp[1]+"\\"+tmp.SubString(4,tmp.Length()-3),
                    true,p_obs);
        }
        for(int i=0;i<non_recursed->Count;++i){
            String tmp=non_recursed->Strings[i];
            if(!tmp.Length())
                continue;
            sweep(destination+tmp[1]+"\\"+tmp.SubString(4,tmp.Length()-3),
                    false,p_obs);
        }
        // third round; remove obsolete directories resulting from round two
        Form1->Caption="Removing obsolete directories...";
        dirHitList=new TStringList(); // create container
        // gather all directories in destination
        sweep(destination,true,p_dirKiller);
        // kill them all.. that'll distract 'em
        for(int i=0;i<dirHitList->Count;++i){
            String tmp=dirHitList->Strings[i];
            while(RemoveDir(tmp)){ // failure indicates dir ain't empty
                tmp=ExtractFilePath(tmp);
                if(tmp.Length()&&*AnsiLastChar(tmp)=='\\')
                    tmp.Delete(tmp.Length(),1);
            }
        }
        delete dirHitList;
    }
    catch(Exception &exception){
        Application->ShowException(&exception);
    }
THE_END:
    delete keep;
    delete include;
    delete exclude;
    delete recursed;
    delete non_recursed;
    ReleaseMutex(mutex);
    return 0;
}
//---------------------------------------------------------------------------
